@@ -10,14 +10,17 @@ import time
 import subprocess
 import urllib
 
+from datetime import datetime
+
 from dulwich.repo import Repo
 from dulwich.errors import NotGitRepository
 
 from tiddlyweb.model.tiddler import Tiddler
+from tiddlyweb.serializer import TiddlerFormatError
 from tiddlyweb.store import StoreLockError, NoBagError, NoTiddlerError
-from tiddlyweb.stores.text import Store as TextStore
-from tiddlyweb.util import LockError, write_lock, write_unlock, \
-        read_utf8_file, write_utf8_file
+from tiddlyweb.stores.text import Store as TextStore, _encode_filename
+from tiddlyweb.util import (binary_tiddler, LockError, write_lock, write_unlock,
+        read_utf8_file, write_utf8_file)
 
 
 class Store(TextStore):
@@ -34,21 +37,27 @@ class Store(TextStore):
 
         try:
             tiddler_files = (filename for filename
-                    in self._files_in_dir(tiddlers_dir))
+                    in self._files_in_dir(tiddlers_dir)
+                    if not filename == '.gitkeep')
         except (IOError, OSError), exc:
             raise NoBagError('unable to list tiddlers in bag "%s": %s' %
                     (bag.name, exc))
 
+        bin_dir = os.path.basename(self._binaries_dir(bag.name))
         for filename in tiddler_files:
-            title = urllib.unquote(filename).decode('utf-8')
-            yield Tiddler(title, bag.name)
+            if not filename == bin_dir:
+                title = urllib.unquote(filename).decode('utf-8')
+                yield Tiddler(title, bag.name)
 
     def list_tiddler_revisions(self, tiddler):
         tiddler_filename = self._tiddler_base_filename(tiddler)
+        if not os.path.isfile(tiddler_filename):
+            raise NoTiddlerError('unable to list revisions for tiddler "%s"'
+                    % tiddler.title)
 
         try:
             relative_path = os.path.relpath(tiddler_filename, start=self._root)
-            revisions = run('git', 'log', '--format=%H', relative_path,
+            revisions = run('git', 'log', '--format=%H', '--', relative_path,
                     cwd=self._root) # TODO: should be handled via Dulwich
         except subprocess.CalledProcessError, exc:
             raise NoTiddlerError('unable to list revisions for tiddler "%s": %s'
@@ -71,10 +80,17 @@ class Store(TextStore):
         revision = run('git', 'log', '-n1', '--format=%H', cwd=self._root) # TODO: should be handled via Dulwich
         tiddler.revision = revision.strip()
 
+        if binary_tiddler(tiddler):
+            with open(self._binary_filename(tiddler), 'rb') as fh:
+                tiddler.text = fh.read() # XXX: is it this simple?
+
         return tiddler
 
     def tiddler_put(self, tiddler):
         tiddler_filename = self._tiddler_base_filename(tiddler)
+        bin_dir = self._binaries_dir(tiddler.bag)
+        if tiddler_filename == bin_dir:
+            raise TiddlerFormatError("reserved tiddler title")
 
         # the following section is copied almost verbatim from the text store
         # TODO: refactor the text store for granular reusability
@@ -104,11 +120,32 @@ class Store(TextStore):
             tiddler.creator = tiddler.modifier
             tiddler.created = tiddler.modified
 
+        binary_data = None
+        if binary_tiddler(tiddler):
+            try:
+                os.mkdir(bin_dir)
+            except OSError, exc: # already exists
+                pass
+            binary_filename = self._binary_filename(tiddler)
+            with open(binary_filename, 'wb') as fh:
+                fh.write(tiddler.text) # XXX: is it this simple?
+            binary_data = tiddler.text
+            # ensure metadata file changes when binary contents change, thus
+            # making it part of any commit - the metadata file is considered the
+            # authoritative source for revisions
+            tiddler.text = '~.~.~bincue~.~.~ %s' % datetime.now()
+
         write_utf8_file(tiddler_filename,
                 self.serializer.serialization.tiddler_as(tiddler))
         write_unlock(tiddler_filename)
 
-        commit_id = self._commit('tiddler put', tiddler_filename) # XXX: message too technical?
+        commit_files = [tiddler_filename]
+
+        if binary_data is not None:
+            tiddler.text = binary_data # restore original
+            commit_files.append(binary_filename)
+
+        commit_id = self._commit('tiddler put', *commit_files) # XXX: message too technical?
         tiddler.revision = commit_id # TODO: use abbreviated commit hash
 
     def tiddler_delete(self, tiddler): # XXX: prone to race condition due to separate Git operation
@@ -117,6 +154,8 @@ class Store(TextStore):
             raise NoTiddlerError('%s not present' % tiddler_filename)
 
         os.remove(tiddler_filename)
+        if binary_tiddler(tiddler):
+            run('git', 'rm', self._binary_filename(tiddler), cwd=self._root) # TODO: should be handled via Dulwich
 
         self._commit('tiddler delete', tiddler_filename) # XXX: message too technical?
 
@@ -162,6 +201,9 @@ class Store(TextStore):
         user_filename = self._user_path(user)
         self._commit('user delete', user_filename) # XXX: message too technical?
 
+    def search(self, search_query):
+        raise NotImplementedError
+
     def _get_tiddler_revision(self, tiddler, tiddler_filename):
         relative_path = os.path.relpath(tiddler_filename, start=self._root)
         target = '%s:%s' % (tiddler.revision, relative_path)
@@ -184,7 +226,7 @@ class Store(TextStore):
         host = '%s:%s' % (host['host'], host['port'])
         if host.endswith(':80'): # TODO: use proper URI parsing instead
             host = host[:-3]
-        user = self.environ['tiddlyweb.usersign']['name']
+        user = self.environ.get('tiddlyweb.usersign', {}).get('name', None)
         author = '%s <%s@%s>' % (user, user, host)
         committer = 'tiddlyweb <tiddlyweb@%s>' % host
 
@@ -198,6 +240,13 @@ class Store(TextStore):
                 os.path.join('tiddlers', '.gitkeep')]
         return [os.path.join(bag_path, filename) for filename in bag_files]
 
+    def _binaries_dir(self, bag_name):
+        return os.path.join(self._tiddlers_dir(bag_name), '_binaries')
+
+    def _binary_filename(self, tiddler):
+        return os.path.join(self._binaries_dir(tiddler.bag),
+                _encode_filename(tiddler.title))
+
 
 def run(cmd, *args, **kwargs):
     """
